%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Matlab code for Bayesian Tensor Factorisations for Time Series of %%%
%%% Counts                                                            %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%% Univariate Cases %%%

clear all;

seed=9;  
rng(seed);  
RandStream.getGlobalStream;

%%% read data
ddata = importdata('xxx.csv');
A = ddata.data;
sA = size(A);

rate_pre = 0.3; %size of pre-training data
rate_train = 0.8; %size of training data

T1 = floor(sA(1)*rate_pre);
T = floor(sA(1)*rate_train);
T2 = T-T1;

data_pretrain = A(1:T1); %pre-training data
data_train_test = A((T1+1):end); %training data + testing data
data_train = A((T1+1):T); %training data excluding pre-training data
data_test = A((T+1):end); %testing data

N = size(data_train_test, 1);

%%% stage 0: use pre-training data to label the dataset by mixture of Poisson distribution 
%%% needed for large counts
D_cate = 10; %total number of categories
[lambda_D, pi_mean] = Pois_Mix_Fun_2(data_pretrain,floor(max(data_pretrain)/2),D_cate,10000);
lambda_DD = lambda_D;
DD_train_test = [];
for i = 1:size(data_train_test,1)
    AA = poisspdf(data_train_test(i), lambda_DD);
    DD_train_test(i) = find(AA == max(AA));
end
DD_train_test = DD_train_test(:);
d0=max(A)+1;

Y0=data_train_test;
Y=data_train;
MMM=tabulate(Y); MMM(:,3);

%%% stage 1: important lags selection
qmax = 10; % maximum dependence order
p=0; % number of external predictors
pnew=(qmax+p);
dnew=max(DD_train_test)*ones(1,qmax);   % redefine the number of levels of each predictor, now the predictor set includes qmax preceding y values
dmax=max(dnew);
pM=zeros(pnew,dmax);
for j=1:pnew
    pM(j,1:dnew(j))=exp(-(j*(1:dnew(j))/2)); % prior probability for k_{j}
    pM(j,1:dnew(j))=pM(j,1:dnew(j))./sum(pM(j,1:dnew(j)));
end
lambdaalpha=max(data_train)/2;   % prior for Gamma for lambda
%lambdaalpha=mode(A);
lambdabeta=1;
pigamma(1:pnew)=1./dnew;     % prior for Dirichlet distribution for pi
pigamma(1:pnew)=1./100;
ep=max(10,floor(log(T)/log(4))); %expected maximum number of predictors
ep=min(pnew,ep);
%ep=6;


fprintf('\n--- First Stage --- \n\n');
N1=2000;     % number of iterations for first stage
np=0;       % number of predictors included in the model 
Nnew=N-qmax;
X0new=zeros(Nnew,pnew);                 %                                     ---- Structure of X0new  ---  
Y0new=Y0((qmax+1):N);                   %
for j=1:qmax                            %   Y(qmax+1)->    Y(qmax)   Y(qmax-1)  ...       Y(1)    X(qmax+1,1)     ...   X(qmax+1,p)
    X0new(:,j)=Y0(qmax+1-j:(N-j));      %   Y(qmqx+2)->    Y(qmax+1) Y(qmax)    ...       Y(2)    X(qmax+2,1)     ...   X(qmax+2,p)
end                                     %     ...    ->      ...      ...       ...       ...         ...         ...       ...

D_X0new=zeros(Nnew,pnew);                 %                                     ---- Structure of X0new  ---  
Y0new=Y0((qmax+1):N);                   %
for j=1:qmax                            %   Y(qmax+1)->    Y(qmax)   Y(qmax-1)  ...       Y(1)    X(qmax+1,1)     ...   X(qmax+1,p)
    D_X0new(:,j)=DD_train_test(qmax+1-j:(N-j));      %   Y(qmqx+2)->    Y(qmax+1) Y(qmax)    ...       Y(2)    X(qmax+2,1)     ...   X(qmax+2,p)
end                                     %     ...    ->      ...      ...       ...       ...         ...         ...       ...


T1new=(T-qmax-T1);
Xnew=D_X0new(1:T1new,:);
%Xnew=X0new(1:T1new,:);
Ynew=Y0new(1:T1new);
M=ones(N1+1,pnew);
G=zeros(pnew,dmax);
for j=1:pnew
    G(j,1:dnew(j))=1;
end
z=ones(T1new,pnew);
log0=zeros(N1,1);
cM=zeros(pnew,dmax);
for j=1:pnew                                % There are (2^r-2)/2 ways to split r objects into two non-empty groups.
    cM(j,1:dnew(j))=(2.^(1:dnew(j))-2)/2;   % (Consider 2 cells and r objects, subtract 2 for the two cases in which 
end                                         % one cell receives all objects, divide by 2 for symmetry.) 


for k=1:N1
    M00=M(k,:);     % initiate M00={k_{1},...,k_{p}}, current values of k_{j}'s for the kth iteration
    for j=1:pnew
        M0=M00(j);  % M0=k_{j}, current value
        if M0==1    % if x_{j} is not included, propose its inclusion OR a switch with an existing important predictor
          % propose the inclusion of x_{j} with randomly generated cluster mappings for different levels of x_{j}
          if np<ep  % propose inclusion of x_{j} only if np<ep so that np never exceeds ep
            new=binornd(1,0.5*ones(1,dnew(j)-1));	% propose new mapping for (d_{j}-1) values at 1
            while sum(new)==0
                new=binornd(1,0.5*ones(1,dnew(j)-1));
            end
            GG=G(j,1:dnew(j))+[0 new]; % keep the first one at 1, propose new cluster mappings for the other (d_{j}-1) levels of x_{j}
            zz=z;                   % zz initiated at z
            zz(:,j)=GG(Xnew(:,j));  % proposed new zz by mapping to new cluster configurations for the observed values of x_{j}
            ind2=find(M00>1);       % current set of relevant predictors
            if isempty(ind2)        % if no predictor is currently important
                ind2=1;
            end
            MM=M00;                 % MM initiated at {k_{1},...,k_{p}}, current values, with k_{j}=1 by the if condition
            MM(j)=2;                % proposed new value of k_{j}=2
            ind1=find(MM>1);        % proposed set of important predictors, now includes x_{j}, since MM(j)=2
            logR=logml_Pois(zz(:,ind1),Ynew,MM(ind1),pM(ind1,:),lambdaalpha,lambdabeta)-logml_Pois(z(:,ind2),Ynew,M00(ind2),pM(ind2,:),lambdaalpha,lambdabeta);
            logR=logR+log(cM(j,dnew(j)));
            if log(rand)<logR
                G(j,1:dnew(j))=GG;
                M00=MM;
                z=zz;
                np=np+1;
            end
          end
        end
        if M0>1&&M0<dnew(j) % if 1<k_{j}<d_{j} (recall that M0=M00(j)=k_{j})
            if rand<0.5     % with prob 0.5 merge two mapped values into one
                cnew=randsample(M0,2);
                lnew=max(cnew);
                snew=min(cnew);
                GG=G(j,1:dnew(j));
                GG(GG==lnew)=snew;  % replace all lnews by snews
                GG(GG==M0)=lnew;    % replace the largest cluster mapping by lnew (lnew itself may equal M0, in which case GG remains unchanged by this move)
                zz=z;               % zz initiated at z
                zz(:,j)=GG(Xnew(:,j)); % proposed new z_{tj} as per the proposed new cluster mappings of the levels of x_{j}
                MM=M00;             % MM initizted at current values {k_{1},...,k_{p}}, with k_{j}=d_{j} by the if condition
                MM(j)=M00(j)-1;     % proposed new value of k_{j}, since two mappings are merged
                ind1=find(MM>1);    % proposed set of important predictors, may not include x_{j} if original k_(j) was at 2
                ind2=find(M00>1);   % current set of important predictors
                if isempty(ind1)
                    ind1=1;
                end
                if isempty(ind2)
                    ind2=1;
                end
                logR=logml_Pois(zz(:,ind1),Ynew,MM(ind1),pM(ind1,:),lambdaalpha,lambdabeta)-logml_Pois(z(:,ind2),Ynew,M00(ind2),pM(ind2,:),lambdaalpha,lambdabeta);
                if M0>2
                    [z0,mm]=unique(sort(GG),'legacy');
                    gn=mm-[0 mm(1:(end-1))];
                    logR=logR-log(sum(cM(j,gn)))+log(M00(j)*(M00(j)-1)/2);
                else
                    logR=logR-log(cM(j,dnew(j))); 
                end
                if log(rand)<logR
                    G(j,1:dnew(j))=GG;
                    M00=MM;
                    z=zz;
                    if M00(j)==1
                        np=np-1;
                    end
                end
            else        % with prob 0.5 split one mapped value into two
                [z0,mm]=unique(sort(G(j,1:dnew(j))),'legacy');                        % z0 are unique cluster mappings, mm contains their positions
                gn=mm-[0 mm(1:(end-1))];            % frequencies of z0
                pgn=cM(j,gn)/sum(cM(j,gn));         % see the definition of cM
                rr=sum(mnrnd(1,pgn).*(1:M00(j)));   % rr is the state to split, gn(rr) is the frequency of rr
                new=binornd(1,0.5*ones(1,gn(rr)-1));% propose new mapping for (gn(rr)-1) values at rr
                while sum(new)==0
                    new=binornd(1,0.5*ones(1,gn(rr)-1));
                end
                GG=G(j,1:dnew(j));
                GG(GG==rr)=rr+(M0+1-rr)*[0 new]; % keep first value at rr, propose new mapping (M0+1) for the rest of (gn(rr)-1) values at rr
                zz=z;               % zz initiated at z
                zz(:,j)=GG(Xnew(:,j)); % proposed new z_{tj} as per the proposed new cluster mappings of the levels of x_{j}
                MM=M00;             % MM initizted at current values {k_{1},...,k_{p}}
                MM(j)=M00(j)+1;     % proposed new value of k_{j}, since one original mapped value is split into two
                ind1=find(MM>1);    % proposed set of important predictors
                ind2=find(M00>1);   % current set of important predictors
                if isempty(ind2)
                    ind2=1;
                end                
                logR=logml_Pois(zz(:,ind1),Ynew,MM(ind1),pM(ind1,:),lambdaalpha,lambdabeta)-logml_Pois(z(:,ind2),Ynew,M00(ind2),pM(ind2,:),lambdaalpha,lambdabeta);
                if M00(j)<dnew(j)-1
                    logR=logR-log(M00(j)*(M00(j)+1)/2)+log(sum(cM(j,gn)));
                else
                    logR=logR-log(dnew(j)*(dnew(j)-1)/2);
                end
                if log(rand)<logR
                    G(j,1:dnew(j))=GG;
                    M00=MM;
                    z=zz;
                end     
            end
        end        
        if M0==dnew(j)              % if M0=k_{j}=d_{j}, propose a merger of two different cluster mappings
            cnew=randsample(dnew(j),2);
            lnew=max(cnew);
            snew=min(cnew);
            GG=G(j,1:dnew(j));
            GG(GG==lnew)=snew;      % replace all lnews by snews
            GG(GG==M0)=lnew;        % replace the largest cluster mapping d_{j} by lnew (lnew itself can be d_{j}, in which case GG remains unchanged by this move)
            zz=z;                   % zz initiated at z
            zz(:,j)=GG(Xnew(:,j));  % proposed new z_{tj} as per the proposed new cluster mappings of the levels of x_{j}
            MM=M00;                 % MM initizted at current values {k_{1},...,k_{p}}, with k_{j}=d_{j} by the if condition
            MM(j)=dnew(j)-1;        % proposed new value of k_{j}, since originally k_{j}=d_{j} and now two mappings are merged
            ind1=find(MM>1);        % proposed set of important predictors, does not include x_{j} when d_(j)=2
            if isempty(ind1)
               ind1=1;
            end
            ind2=find(M00>1);       % current set of important predictors
            if isempty(ind2)
               ind2=1;
            end            
            logR=logml_Pois(zz(:,ind1),Ynew,MM(ind1),pM(ind1,:),lambdaalpha,lambdabeta)-logml_Pois(z(:,ind2),Ynew,M00(ind2),pM(ind2,:),lambdaalpha,lambdabeta);
            logR=logR+log(dnew(j)*(dnew(j)-1)/2);
            if log(rand)<logR
                G(j,1:dnew(j))=GG;
                M00=MM;
                z=zz;
            end   
            if M00(j)==1
                np=np-1;
            end
        end 
    end
    M(k+1,:)=M00;
    
    % print informations in each iteration
    ind1=find(M00>1);
    if isempty(ind1)
        ind1=1;
    end   
    %log0(k)=logml_Pois(z(:,ind1),Ynew,M(k+1,ind1),pM(ind1,:),lambdaalpha,lambdabeta);
    [aaY,bY]=find(M(k+1,1:qmax)-1);
    [aaX,bX]=find(M(k+1,(qmax+1):pnew)-1);
    fprintf('k = %i, %i important predictors = {',k,np);
    for i=1:length(bY)
        fprintf(' Y(t-%i)(%i)',bY(i),M(k+1,bY(i)));
    end
    for i=1:length(bX)
        fprintf(' X%i(%i)',bX(i),M(k+1,qmax+bX(i)));
    end
    fprintf(' }. %f \n',log0(k));

end

VarSelect=(M(1:N1,:)>1);
VarSelectProps=sum(VarSelect(floor(N1/2)+1:N1,:),1)./(N1-floor(N1/2));
display(VarSelectProps);  

%%% stage 2: posteriors of random variables
fprintf('\n --- Second Stage --- \n\n');
pause(1);
simsize=2000;            % number of iterations for second stage
burnin=1000;             %floor(simsize/5);
log01=zeros(simsize,1);
log02=zeros(simsize,1);
nord=zeros(simsize,1);
npred=zeros(simsize,1);
npredact=zeros(simsize,1);
nzstar=zeros(simsize,1);

% Storage for prediction
gap=20;
kgap=1;
D_Xfut=zeros(N-T2,qmax);
for qq=1:qmax
	D_Xfut(:,qq) = DD_train_test((T2+1-qq):(N-qq));
end
PP=zeros(floor((simsize-burnin)/gap),(N-T2));

ind00=find(VarSelectProps>0.5);   % selected predictors for which k_{j}>1
if isempty(ind00)
    ind00=1;
end
M0=M00(ind00);       % k_{j}'s for the selected predictors
p00=length(ind00);   % number of selected predictors
z00=z(:,ind00);
x0=Xnew(:,ind00);    % selected predictor values
d00=dnew(ind00);
dmax=max(d00);       % redefine dmax
pigamma00=pigamma(ind00);
pi=zeros(p00,dmax,dmax);% pi(j,x_{j},k,iteration no), x_{j} runs in {1,...,d_{j}} and k runs in {1,...,k_{j}}, dmax is used for d_{j} and k_{j}, extra values will not be used or updated
    
kstar=100;
pigammastar=1;
Vstar=betarnd(ones(1,kstar),pigammastar*ones(1,kstar));
oneminusVstarProd=[1,cumprod(1-Vstar)];
oneminusVstarProd=oneminusVstarProd(1:kstar);
pistar=oneminusVstarProd.*Vstar;
betabeta = 1;
betaalpha = max(A)/2;
lambdaast  = gamrnd(betaalpha,betabeta,1,kstar);

%MCMC diagnostics of predicton
log_socre = [];
k = 1;
 while k< simsize+1
      try
 % pi
    cp=zeros(p00,dmax,dmax);   % counts for pi,first=j,second=value of x,third=value of z
    for t=1:T1new
        for j=1:p00
            cp(j,x0(t,j),z00(t,j))=cp(j,x0(t,j),z00(t,j))+1;
        end
    end
    for j=1:p00
        for s=1:d00(j)
            rr=gamrnd(cp(j,s,1:M0(j))+pigamma00(j),1);
            pi(j,s,1:M0(j))=rr/sum(rr);	% s=x_{j}, only the first k_{j} elements are updated, rest are left at zero
        end
        % switch labels 
        % in the reshaped d_{j}*k_{j} matrix below, the rows correspond to the values of x_{j} and the columns to the values of h_{j}, the sum across each row is one 
        [~, qq2]=sort(sum(reshape(pi(j,:,1:M0(j)),dmax,M0(j)),1),'descend');
        for s=1:d00(j)
            pi(j,s,1:M0(j))=pi(j,s,qq2);    % column labels h_{j}'s are switched
        end
        for t=1:T1new
            z00(t,j)=find(qq2==z00(t,j));       % z00_{t,j}'s are switched
        end
    end

clT=tensor(zeros([d0,M0]),[d0,M0]);         % d0 levels of the response y, M0={k_{1},..,k_{p00}} soft clustered levels of x_{1},...,x_{p00}
[z0,m]=unique(sortrows([Ynew+1 z00]),'rows','legacy'); % z0 are the sorted unique combinations of (y,z_{1},...,z_{p00}), m contains their positions on {1,...,T}
clT(z0)=clT(z0)+m-[0;m(1:(end-1))];         % add the differences in positions to cells of clT corresponding to the unique combinations -> gives the number of times (y,z_{1},...,z_{p00}) appears 
clTdata=tenmat(clT,1);                      % matrix representation of the tensor clT, with rows of the matrix corresponding to dimension 1 i.e. the levels of y
sz=size(clTdata);

bb = 0:(d0-1);
bb0 = clTdata.data.*bb';
nh1tohqy = sum(bb0, 1);
nh1tohq = sum(clTdata.data, 1);

% zstar
zstarmat_01=zeros(sz(2),kstar);
for jj=1:sz(2)
pizhstar = log(pistar) + log(lambdaast).*(repmat(nh1tohqy(jj),1,kstar)) - lambdaast.*(repmat(nh1tohq(jj),1,kstar));
pizhstar = exp(pizhstar);
pizhstar = pizhstar/sum(pizhstar);
if isnan(sum(pizhstar))
            pizhstar=(1/kstar)*ones(1,kstar);
end
zstarmat_01(jj,:)=mnrnd(1,pizhstar,1);
end
nstar_01=sum(zstarmat_01,1);
zstar_01=sum(bsxfun(@times,zstarmat_01,1:kstar),2)';

% pistar
Vstar=betarnd(1+nstar_01,pigammastar+sum(nstar_01)-cumsum(nstar_01));
oneminusVstarProd=[1,cumprod(1-Vstar)];
oneminusVstarProd=oneminusVstarProd(1:kstar);
pistar=oneminusVstarProd.*Vstar;

 % lambdastar
astar_01=nh1tohq*zstarmat_01;
astar_02=nh1tohqy*zstarmat_01;
lambdastarmat_01=gamrnd(astar_02+betaalpha, 1./(astar_01+betabeta));

% lambda
lambdamat_01=lambdastarmat_01(:,zstar_01);
%lambdamat_01= sort(lambdamat_01);
lambda_01=tensor(lambdamat_01, M0);


% z
if p00 == 1
    qq_00 = zeros(T1new,M0(j)); % for computing p(z=h|-)
     for h=1:M0
         lll = lambda_01([z00(:,1:(j-1)), h*ones(T1new,1), z00(:,(j+1):p00)]);
         lll = lll.data;
            aaa = poisspdf(Ynew, lll);
            qq_00(:,h)=pi(j,z00(:,j),h).*(reshape(double(aaa),T1new,1))';
     end
        qq_00=bsxfun(@rdivide,qq_00,(sum(qq_00,2)));                 % normalize qq
        qq_00(isnan(qq_00)) = 1;
        qq_00=bsxfun(@rdivide,qq_00,(sum(qq_00,2))); 
        z00=sum(bsxfun(@times,mnrnd(1,qq_00),1:M0),2); % sample from multinomial(1,qq) and then multiply with {1,...,k_{j}} 
end

if p00 > 1
    for j=1:p00
        qq_01=zeros(T1new,M0(j)); % for computing p(z=h|-)
        for h=1:M0(j)
            %aaa = lambda_01([z00(:,1:(j-1)), h*ones(Tnew,1), z00(:,(j+1):p00)]).^Ynew.*exp(-lambda_01([z00(:,1:(j-1)), h*ones(Tnew,1), z00(:,(j+1):p00)]));
            aaa = poisspdf(Ynew, lambda_01([z00(:,1:(j-1)), h*ones(T1new,1), z00(:,(j+1):p00)]));
            qq_01(:,h)=pi(j,x0(:,j),h).*(reshape(double(aaa),T1new,1))';
        end
        qq_01=bsxfun(@rdivide,qq_01,(sum(qq_01,2)));                 % normalize qq
        qq_01(isnan(qq_01)) = 1;
        qq_01=bsxfun(@rdivide,qq_01,(sum(qq_01,2)));  
        z00(:,j)=sum(bsxfun(@times,mnrnd(1,qq_01),1:M0(j)),2); % sample from multinomial(1,qq) and then multiply with {1,...,k_{j}} 
    end   
end


     Mact=ones(1,pnew);
    for j=1:p00
        Mact(j)=length(unique(z00(:,j)));
    end
    npredact(k)=length(find(Mact>1));
    if npredact(k)>0
        nord(k)=find(Mact>1,1,'last');
    end
    nzstar(k)=length(unique(zstar_01));

    
    % print informations in each iteration
%     log01(k)=logml(z00,Ynew,M00(ind00),pM(ind00,:),lambdaalpha);
%     log02(k)=logmly(z00,Ynew,M00(ind00),lambdaalpha);
%     
    fprintf('k = %i, %i important predictors = {',k,p00);
    for i=1:p00
        fprintf(' Y(t-%i)(%i)',ind00(i),M0(i));
    end
    fprintf(' } \n'); 
    
  
    
    %prediction
    if k>burnin && mod(k,gap)==0
    
    U=cell({1:p00});
    for j=1:p00
        U{j}=reshape(pi(j,1:dmax,1:M0(j)),dmax,M0(j));
    end
    
    PPE=tensor(ttensor(tensor(double(lambda_01), M0),U));
    if p00 == 1
         PPEE = PPE(D_Xfut(:,ind00));
         PPEE = PPEE.data;
         PP(kgap,:)=PPEE';
    end
    
    if p00 > 1
    PPE=tensor(ttensor(tensor(double(lambda_01), M0),U));
    PP(kgap,:)=reshape(PPE(D_Xfut(:,ind00)), 1, N-T2);
    end
    
    %%%%%%%%%%-log predictive score%%%%%%%%
    ypred=Y0((T2+1):end)';
ypred_unique = unique(ypred);
log_s_ten = [];
for uni_i = ypred_unique
lst = poisspdf(uni_i, double(lambda_01));
log_s_ten{uni_i+1} = tensor(ttensor(tensor(lst, M0),U)); % +1 since no 0 index for cell
end

for  i = 1: length(ypred)
        log_s = log_s_ten{ypred(i)+1}(D_Xfut(i,ind00));
        log_score(i,kgap) = log_s;
end

% log_score = 0;
% for  i = 1: length(ypred)
%         log_s = log_s_ten{ypred(i)+1}(D_Xfut(i,ind00));
%         log_s = - log(log_s);
%         log_score = log_score + log_s;
% end
%     log_pred_score(kgap) = log_score;
    %%%%%%%%%%%%%%%%%%%%%
    
    %lambda_mat(:,kgap) = double(lambda_01);
    %U1_mat(:,:,kgap) = U{1};
    %U2_mat(:,:,kgap) = U{2};
    
%     quantile_l = []
%     %%%%%%%Quantile%%%%%%
%     prob_den_table = []
% for  i = 1: length(ypred)
%         lllst = []
%         for num = 1:500
%             llst = poisspdf(num, double(lambda_01));
%             llst = tensor(ttensor(tensor(llst, M0),U));
%             llst = llst(D_Xfut(i,ind00));
%             lllst = [lllst llst];
%         end
%      prob_den_table(:,i) = lllst;         
% end
% 
% prob_den_table_all(:,:,kgap)=prob_den_table;
    
    
    
    
    %%%%%%%%%%%%%%%%%%%%%%
    
    
  
    kgap=kgap+1;

    end
    
       catch
     k = k-3;
     warning('Something happened');
     %k = k-1;
      end
    
 k = k+1;
    
 end
lps = mean(-log(mean(log_score,2)))





%%%Multivariate Cases%%%
clear all;

seed=9;  
rng(seed);  
RandStream.getGlobalStream; 
 
 
%%% read data
ddata = importdata('xxx.csv');

A = ddata.data;
sA = size(A);

rate_pre = 0.3;
rate_train = 0.9;
T = floor(sA(1)*rate_train);
T1 = floor(sA(1)*rate_pre);

data_pretrain = A(1:T1,:);
data_train_test = A((T1+1):end,:);
data_train = A((T1+1):T,:);
data_test = A((T+1):end,:);
N = size(data_train_test, 1);

%%% stage 0: use pre-training data to label the dataset by mixture of Poisson distribution 
%%% needed for large counts
DD_train_test = [];
for i = 1:sA(2)
    D = 6;
    [lambda_D, ~] = Pois_Mix_Fun_2(data_pretrain(:,i),floor(max(data_pretrain(:,i))/2),D,1000);

    lambda_DD = lambda_D;
    
    for j = 1:size(data_train_test,1)
    AA = poisspdf(data_train_test(j,i), lambda_DD);
    DD_train_test(j,i) = find(AA == max(AA));
    end
end

T2 = T-T1;
target = 5;

%%%Approximate Sampler%%%
nom = size(data_train_test,2); 
A_target = data_train_test(:,target);
A_cova = DD_train_test;
A_cova(:,target) = [];

d0=max(data00(1:T, target))+1;

y = A_target;
train=1:T2;
Y0=y;
Y=y(train);

%%% stage 1: important lags selection
qmax=10;     % maximum dependence order
p=0;         % number of external predictors
pnew=(qmax+p);
dnew=max(DD_train_test,[],'all')*ones(1,qmax);   % redefine the number of levels of each predictor, now the predictor set includes qmax preceding y values
dnew = repmat(dnew, 1, nom);
dmax=max(dnew);
pM=zeros(pnew,dmax);
for j=1:pnew
    pM(j,1:dnew(j))=exp(-(j*(1:dnew(j))/2)); % prior probability for k_{j}
    pM(j,1:dnew(j))=pM(j,1:dnew(j))./sum(pM(j,1:dnew(j)));
end
pM = repmat(pM, nom, 1);

poisalpha = floor(max(Y)/2); % prior for Poisson distribution for lambda
poisbeta = 1;

pigamma(1:pnew*nom)=1./dnew;     % prior for Dirichlet distribution for pi
ep=max(10,floor(log(T)/log(4))); %expected maximum number of predictors
ep=min(pnew,ep);
%ep = 6;


fprintf('\n--- First Stage --- \n\n');
N1=1000;     % number of iterations for first stage
np=0;       % number of predictors included in the model 
Nnew=N-qmax;
X0new=zeros(Nnew,pnew*nom);                 %                                     ---- Structure of X0new  ---  
Y0new=Y0((qmax+1):N);                   %
for j=1:qmax                            %   Y(qmax+1)->    Y(qmax)   Y(qmax-1)  ...       Y(1)    X(qmax+1,1)     ...   X(qmax+1,p)
    X0new(:,j)=DD_train_test(qmax+1-j:(N-j), target);      %   Y(qmqx+2)->    Y(qmax+1) Y(qmax)    ...       Y(2)    X(qmax+2,1)     ...   X(qmax+2,p)
end                                     %     ...    ->      ...      ...       ...       ...         ...         ...       ...

for i = 1:(nom-1)
    for j = 1:qmax
        X0new(:,i*qmax+j) = A_cova(qmax+1-j:(N-j), i);
    end
end

T2new=(T2-qmax);
Xnew=X0new(1:T2new,:);
Ynew=Y0new(1:T2new);

M=ones(N1+1,pnew*nom);
G=zeros(pnew*nom,dmax);
for j=1:pnew*nom
    G(j,1:dnew(j))=1;
end
z=ones(T2new,pnew*nom);
log0=zeros(N1,1);
cM=zeros(pnew*nom,dmax);
for j=1:pnew*nom                                % There are (2^r-2)/2 ways to split r objects into two non-empty groups.
    cM(j,1:dnew(j))=(2.^(1:dnew(j))-2)/2;   % (Consider 2 cells and r objects, subtract 2 for the two cases in which 
end    

for k=1:N1
    M00=M(k,:);     % initiate M00={k_{1},...,k_{p}}, current values of k_{j}'s for the kth iteration
    for j=1:pnew*nom
        M0=M00(j);  % M0=k_{j}, current value
        if M0==1    % if x_{j} is not included, propose its inclusion OR a switch with an existing important predictor
          % propose the inclusion of x_{j} with randomly generated cluster mappings for different levels of x_{j}
          if np<ep  % propose inclusion of x_{j} only if np<ep so that np never exceeds ep
            new=binornd(1,0.5*ones(1,dnew(j)-1));	% propose new mapping for (d_{j}-1) values at 1
            while sum(new)==0
                new=binornd(1,0.5*ones(1,dnew(j)-1));
            end
            GG=G(j,1:dnew(j))+[0 new]; % keep the first one at 1, propose new cluster mappings for the other (d_{j}-1) levels of x_{j}
            zz=z;                   % zz initiated at z
            zz(:,j)=GG(Xnew(:,j));  % proposed new zz by mapping to new cluster configurations for the observed values of x_{j}
            ind2=find(M00>1);       % current set of relevant predictors
            if isempty(ind2)        % if no predictor is currently important
                ind2=1;
            end
            MM=M00;                 % MM initiated at {k_{1},...,k_{p}}, current values, with k_{j}=1 by the if condition
            MM(j)=2;                % proposed new value of k_{j}=2
            ind1=find(MM>1);        % proposed set of important predictors, now includes x_{j}, since MM(j)=2
            logR=logml_Pois(zz(:,ind1),Ynew,MM(ind1),pM(ind1,:),poisalpha,poisbeta)-logml_Pois(z(:,ind2),Ynew,M00(ind2),pM(ind2,:),poisalpha, poisbeta);
            logR=logR+log(cM(j,dnew(j)));
            if log(rand)<logR
                G(j,1:dnew(j))=GG;
                M00=MM;
                z=zz;
                np=np+1;
            end
          end
        end
        if M0>1&&M0<dnew(j) % if 1<k_{j}<d_{j} (recall that M0=M00(j)=k_{j})
            if rand<0.5     % with prob 0.5 merge two mapped values into one
                cnew=randsample(M0,2);
                lnew=max(cnew);
                snew=min(cnew);
                GG=G(j,1:dnew(j));
                GG(GG==lnew)=snew;  % replace all lnews by snews
                GG(GG==M0)=lnew;    % replace the largest cluster mapping by lnew (lnew itself may equal M0, in which case GG remains unchanged by this move)
                zz=z;               % zz initiated at z
                zz(:,j)=GG(Xnew(:,j)); % proposed new z_{tj} as per the proposed new cluster mappings of the levels of x_{j}
                MM=M00;             % MM initizted at current values {k_{1},...,k_{p}}, with k_{j}=d_{j} by the if condition
                MM(j)=M00(j)-1;     % proposed new value of k_{j}, since two mappings are merged
                ind1=find(MM>1);    % proposed set of important predictors, may not include x_{j} if original k_(j) was at 2
                ind2=find(M00>1);   % current set of important predictors
                if isempty(ind1)
                    ind1=1;
                end
                if isempty(ind2)
                    ind2=1;
                end
                logR=logml_Pois(zz(:,ind1),Ynew,MM(ind1),pM(ind1,:),poisalpha, poisbeta)-logml_Pois(z(:,ind2),Ynew,M00(ind2),pM(ind2,:),poisalpha,poisbeta);
                if M0>2
                    [z0,mm]=unique(sort(GG),'legacy');
                    gn=mm-[0 mm(1:(end-1))];
                    logR=logR-log(sum(cM(j,gn)))+log(M00(j)*(M00(j)-1)/2);
                else
                    logR=logR-log(cM(j,dnew(j))); 
                end
                if log(rand)<logR
                    G(j,1:dnew(j))=GG;
                    M00=MM;
                    z=zz;
                    if M00(j)==1
                        np=np-1;
                    end
                end
            else        % with prob 0.5 split one mapped value into two
                [z0,mm]=unique(sort(G(j,1:dnew(j))),'legacy');                        % z0 are unique cluster mappings, mm contains their positions
                gn=mm-[0 mm(1:(end-1))];            % frequencies of z0
                pgn=cM(j,gn)/sum(cM(j,gn));         % see the definition of cM
                rr=sum(mnrnd(1,pgn).*(1:M00(j)));   % rr is the state to split, gn(rr) is the frequency of rr
                new=binornd(1,0.5*ones(1,gn(rr)-1));% propose new mapping for (gn(rr)-1) values at rr
                while sum(new)==0
                    new=binornd(1,0.5*ones(1,gn(rr)-1));
                end
                GG=G(j,1:dnew(j));
                GG(GG==rr)=rr+(M0+1-rr)*[0 new]; % keep first value at rr, propose new mapping (M0+1) for the rest of (gn(rr)-1) values at rr
                zz=z;               % zz initiated at z
                zz(:,j)=GG(Xnew(:,j)); % proposed new z_{tj} as per the proposed new cluster mappings of the levels of x_{j}
                MM=M00;             % MM initizted at current values {k_{1},...,k_{p}}
                MM(j)=M00(j)+1;     % proposed new value of k_{j}, since one original mapped value is split into two
                ind1=find(MM>1);    % proposed set of important predictors
                ind2=find(M00>1);   % current set of important predictors
                if isempty(ind2)
                    ind2=1;
                end                
                logR=logml_Pois(zz(:,ind1),Ynew,MM(ind1),pM(ind1,:),poisalpha,poisbeta)-logml_Pois(z(:,ind2),Ynew,M00(ind2),pM(ind2,:),poisalpha,poisbeta);
                if M00(j)<dnew(j)-1
                    logR=logR-log(M00(j)*(M00(j)+1)/2)+log(sum(cM(j,gn)));
                else
                    logR=logR-log(dnew(j)*(dnew(j)-1)/2);
                end
                if log(rand)<logR
                    G(j,1:dnew(j))=GG;
                    M00=MM;
                    z=zz;
                end     
            end
        end        
        if M0==dnew(j)              % if M0=k_{j}=d_{j}, propose a merger of two different cluster mappings
            cnew=randsample(dnew(j),2);
            lnew=max(cnew);
            snew=min(cnew);
            GG=G(j,1:dnew(j));
            GG(GG==lnew)=snew;      % replace all lnews by snews
            GG(GG==M0)=lnew;        % replace the largest cluster mapping d_{j} by lnew (lnew itself can be d_{j}, in which case GG remains unchanged by this move)
            zz=z;                   % zz initiated at z
            zz(:,j)=GG(Xnew(:,j));  % proposed new z_{tj} as per the proposed new cluster mappings of the levels of x_{j}
            MM=M00;                 % MM initizted at current values {k_{1},...,k_{p}}, with k_{j}=d_{j} by the if condition
            MM(j)=dnew(j)-1;        % proposed new value of k_{j}, since originally k_{j}=d_{j} and now two mappings are merged
            ind1=find(MM>1);        % proposed set of important predictors, does not include x_{j} when d_(j)=2
            if isempty(ind1)
               ind1=1;
            end
            ind2=find(M00>1);       % current set of important predictors
            if isempty(ind2)
               ind2=1;
            end            
            logR=logml_Pois(zz(:,ind1),Ynew,MM(ind1),pM(ind1,:),poisalpha,poisbeta)-logml_Pois(z(:,ind2),Ynew,M00(ind2),pM(ind2,:),poisalpha,poisbeta);
            logR=logR+log(dnew(j)*(dnew(j)-1)/2);
            if log(rand)<logR
                G(j,1:dnew(j))=GG;
                M00=MM;
                z=zz;
            end   
            if M00(j)==1
                np=np-1;
            end
        end 
    end
    M(k+1,:)=M00;
    
    % print informations in each iteration
    ind1=find(M00>1);
    if isempty(ind1)
        ind1=1;
    end   
    log0(k)=logml_Pois(z(:,ind1),Ynew,M(k+1,ind1),pM(ind1,:),poisalpha,poisbeta);
    [aaY,bY]=find(M(k+1,1:qmax*nom)-1);
    [aaX,bX]=find(M(k+1,(qmax+1):pnew*nom)-1);
    fprintf('k = %i, %i important predictors = {',k,np);
%     for i=1:length(bY)
%         fprintf(' Y(t-%i)(%i)',bY(i),M(k+1,bY(i)));
%     end
    
    for i=1:length(bY)
        fprintf(' Y(%i, t-%i)(%i)', (bY(i)-mod(bY(i), qmax))/qmax + 1, mod(bY(i), qmax) ,M(k+1,bY(i)));
    end
%     for i=1:length(bX)
%         fprintf(' X%i(%i)',bX(i),M(k+1,qmax+bX(i)));
%     end
    fprintf(' }. %f \n',log0(k));
end

VarSelect=(M(1:N1,:)>1);
VarSelectProps=sum(VarSelect(floor(N1/2)+1:N1,:),1)./(N1-floor(N1/2));
display(VarSelectProps);  

%%% stage 2: posteriors of random variables
fprintf('\n --- Second Stage --- \n\n');
pause(1);
simsize=3000;            % number of iterations for second stage
burnin=2000;             %floor(simsize/5);
log01=zeros(simsize,1);
log02=zeros(simsize,1);
nord=zeros(simsize,1);
npred=zeros(simsize,1);
npredact=zeros(simsize,1);
nzstar=zeros(simsize,1);

% Storage for prediction
gap=10;
kgap=1;

D_Xfut=zeros(N-T2,qmax*nom);
for qq=1:qmax
	D_Xfut(:,qq) = DD_train_test((T2+1-qq):(N-qq),target);
end

for i = 1:(nom-1)
    for j = 1:qmax
        D_Xfut(:,i*qmax+j) = A_cova(T2+1-j:(N-j), i);
    end
end
PP=NaN(floor((simsize-burnin)/gap),(N-T2));

ind00=find(VarSelectProps>0.5);   % selected predictors for which k_{j}>1
if isempty(ind00)
    ind00=1;
end

M0=M00(ind00);       % k_{j}'s for the selected predictors
p00=length(ind00);   % number of selected predictors
z00=z(:,ind00);
x0=Xnew(:,ind00);    % selected predictor values
d00=dnew(ind00);
dmax=max(d00);       % redefine dmax
pigamma00=pigamma(ind00);
pi=zeros(p00,dmax,dmax);% pi(j,x_{j},k,iteration no), x_{j} runs in {1,...,d_{j}} and k runs in {1,...,k_{j}}, dmax is used for d_{j} and k_{j}, extra values will not be used or updated
    
kstar=1000;
pigammastar=1;
Vstar=betarnd(ones(1,kstar),pigammastar*ones(1,kstar));
oneminusVstarProd=[1,cumprod(1-Vstar)];
oneminusVstarProd=oneminusVstarProd(1:kstar);
pistar=oneminusVstarProd.*Vstar;
betaalpha = max(Y)/2;
betabeta = 1;
lambdaast  = gamrnd(betaalpha,betabeta,1,kstar);

k = 1;
 while k< simsize+1
      try
 % pi
    cp=zeros(p00,dmax,dmax);   % counts for pi,first=j,second=value of x,third=value of z
    for t=1:T2new
        for j=1:p00
            cp(j,x0(t,j),z00(t,j))=cp(j,x0(t,j),z00(t,j))+1;
        end
    end
    for j=1:p00
        for s=1:d00(j)
            rr=gamrnd(cp(j,s,1:M0(j))+pigamma00(j),1);
            pi(j,s,1:M0(j))=rr/sum(rr);	% s=x_{j}, only the first k_{j} elements are updated, rest are left at zero
        end
        % switch labels 
        % in the reshaped d_{j}*k_{j} matrix below, the rows correspond to the values of x_{j} and the columns to the values of h_{j}, the sum across each row is one 
        [~, qq2]=sort(sum(reshape(pi(j,:,1:M0(j)),dmax,M0(j)),1),'descend');
        for s=1:d00(j)
            pi(j,s,1:M0(j))=pi(j,s,qq2);    % column labels h_{j}'s are switched
        end
        for t=1:T2new
            z00(t,j)=find(qq2==z00(t,j));       % z00_{t,j}'s are switched
        end
    end

clT=tensor(zeros([d0,M0]),[d0,M0]);         % d0 levels of the response y, M0={k_{1},..,k_{p00}} soft clustered levels of x_{1},...,x_{p00}
[z0,m]=unique(sortrows([Ynew+1 z00]),'rows','legacy'); % z0 are the sorted unique combinations of (y,z_{1},...,z_{p00}), m contains their positions on {1,...,T}
clT(z0)=clT(z0)+m-[0;m(1:(end-1))];         % add the differences in positions to cells of clT corresponding to the unique combinations -> gives the number of times (y,z_{1},...,z_{p00}) appears 
clTdata=tenmat(clT,1);                      % matrix representation of the tensor clT, with rows of the matrix corresponding to dimension 1 i.e. the levels of y
sz=size(clTdata);

bb = 0:(d0-1);
bb0 = clTdata.data.*bb';
nh1tohqy = sum(bb0, 1);
nh1tohq = sum(clTdata.data, 1);

% zstar
zstarmat_01=zeros(sz(2),kstar);
for jj=1:sz(2)
pizhstar = log(pistar) + log(lambdaast).*(repmat(nh1tohqy(jj),1,kstar)) - lambdaast.*(repmat(nh1tohq(jj),1,kstar));
pizhstar = exp(pizhstar);
pizhstar = pizhstar/sum(pizhstar);
if isnan(sum(pizhstar))
            pizhstar=(1/kstar)*ones(1,kstar);
end
zstarmat_01(jj,:)=mnrnd(1,pizhstar,1);
end
nstar_01=sum(zstarmat_01,1);
zstar_01=sum(bsxfun(@times,zstarmat_01,1:kstar),2)';

% pistar
Vstar=betarnd(1+nstar_01,pigammastar+sum(nstar_01)-cumsum(nstar_01));
oneminusVstarProd=[1,cumprod(1-Vstar)];
oneminusVstarProd=oneminusVstarProd(1:kstar);
pistar=oneminusVstarProd.*Vstar;

 % lambdastar
astar_01=nh1tohq*zstarmat_01;
astar_02=nh1tohqy*zstarmat_01;
lambdastarmat_01=gamrnd(astar_02+betaalpha, 1./(astar_01+betabeta));

% lambda
lambdamat_01=lambdastarmat_01(:,zstar_01);
lambda_01=tensor(lambdamat_01, M0);


% z
if p00 == 1
    qq_00 = zeros(T2new,M0(j)); % for computing p(z=h|-)
     for h=1:M0
         lll = lambda_01([z00(:,1:(j-1)), h*ones(T2new,1), z00(:,(j+1):p00)]);
         lll = lll.data;
            aaa = poisspdf(Ynew, lll);
            qq_00(:,h)=pi(j,z00(:,j),h).*(reshape(double(aaa),T2new,1))';
     end
        qq_00=bsxfun(@rdivide,qq_00,(sum(qq_00,2)));                 % normalize qq
        qq_00(isnan(qq_00)) = 1;
        qq_00=bsxfun(@rdivide,qq_00,(sum(qq_00,2))); 
        z00=sum(bsxfun(@times,mnrnd(1,qq_00),1:M0),2); % sample from multinomial(1,qq) and then multiply with {1,...,k_{j}} 
end

if p00 > 1
    for j=1:p00
        qq_01=zeros(T2new,M0(j)); % for computing p(z=h|-)
        for h=1:M0(j)
            %aaa = lambda_01([z00(:,1:(j-1)), h*ones(Tnew,1), z00(:,(j+1):p00)]).^Ynew.*exp(-lambda_01([z00(:,1:(j-1)), h*ones(Tnew,1), z00(:,(j+1):p00)]));
            aaa = poisspdf(Ynew, lambda_01([z00(:,1:(j-1)), h*ones(T2new,1), z00(:,(j+1):p00)]));
            qq_01(:,h)=pi(j,x0(:,j),h).*(reshape(double(aaa),T2new,1))';
        end
        qq_01=bsxfun(@rdivide,qq_01,(sum(qq_01,2)));                 % normalize qq
        qq_01(isnan(qq_01)) = 1;
        qq_01=bsxfun(@rdivide,qq_01,(sum(qq_01,2)));  
        z00(:,j)=sum(bsxfun(@times,mnrnd(1,qq_01),1:M0(j)),2); % sample from multinomial(1,qq) and then multiply with {1,...,k_{j}} 
    end   
end


     Mact=ones(1,pnew);
    for j=1:p00
        Mact(j)=length(unique(z00(:,j)));
    end
    npredact(k)=length(find(Mact>1));
    if npredact(k)>0
        nord(k)=find(Mact>1,1,'last');
    end
    nzstar(k)=length(unique(zstar_01));

    
    % print informations in each iteration
%     log01(k)=logml(z00,Ynew,M00(ind00),pM(ind00,:),lambdaalpha);
%     log02(k)=logmly(z00,Ynew,M00(ind00),lambdaalpha);
%     
    fprintf('k = %i, %i important predictors = {',k,p00);
    for i=1:p00
        fprintf(' Y(t-%i)(%i)',ind00(i),M0(i));
    end
    fprintf(' } \n'); 
    
  
    
    %prediction
    if k>burnin && mod(k,gap)==0
    
    U=cell({1:p00});
    for j=1:p00
        U{j}=reshape(pi(j,1:dmax,1:M0(j)),dmax,M0(j));
    end
    
    PPE=tensor(ttensor(tensor(double(lambda_01), M0),U));
    if p00 == 1
         PPEE = PPE(D_Xfut(:,ind00));
         PPEE = PPEE.data;
         PP(kgap,:)=PPEE';
    end
    
    if p00 > 1
    PPE=tensor(ttensor(tensor(double(lambda_01), M0),U));
    PP(kgap,:)=reshape(PPE(D_Xfut(:,ind00)), 1, N-T2);
    end
    
    %%%%%%%%%%-log predictive score%%%%%%%%
    ypred=Y0((T2+1):end)';
ypred_unique = unique(ypred);
log_s_ten = [];
for uni_i = ypred_unique
lst = poisspdf(uni_i, double(lambda_01));
log_s_ten{uni_i+1} = tensor(ttensor(tensor(lst, M0),U)); % +1 since no 0 index for cell
end

for  i = 1: length(ypred)
        log_s = log_s_ten{ypred(i)+1}(D_Xfut(i,ind00));
        log_score(i,kgap) = log_s;
end


% log_score = 0;
% for  i = 1: length(ypred)
%         log_s = log_s_ten{ypred(i)+1}(D_Xfut(i,ind00));
%         log_s = - log(log_s);
%         if log_s > 500
%             log_s = 500;
%         end
%         log_score = log_score + log_s;
% end
%     log_pred_score(kgap) = log_score;
    %%%%%%%%%%%%%%%%%%%%%
    
    
    
  
    kgap=kgap+1;

    end
    
       catch
     k = k-1;
     warning('Something happened');
     %k = k-1;
      end
    
 k = k+1;
    
 end
lps = mean(-log(mean(log_score,2)));


 
 
 